' Loads the .NET framework into your process and instantiates classes in assembly files.
' Use DOTNET_Initialize() to load the .NET framework, this function returns a value <> 0 on success.
' On success use the globally present DOTNET object to create instances of your assemblies.
' Do not try to release this object to load another .NET version or such, it can't be done (using the default domain as is used here).

' Example:

'    Local hr As Long
'    Local o As Dispatch
'    Local v As Variant
'
'    If DOTNET_Initialize( "v2.0.50727", "" ) Then
'
'        hr = DOTNET.CreateObjectByFile( "c:\myassembly.dll", "Class1", o )
'        If hr = 0 Then
'
'            Object Call o.GetDateTimeStr() To v
'
'            MsgBox Format$( hr ) & ", " & Format$( IsObject( o ) ) & $CrLf _
'                & Variant$( v ) & $CrLf _
'                & DOTNET.GetCORVersion & $CrLf _
'                & DOTNET.GetCORSystemDirectory
'
'        End If
'
'    End If


'The following methods are present in the DOTNET interface:

' Returns the interface which is similar to ICorRuntimeHost.
' Property Get RunTimeHost() As DOTNET_ICorRuntimeHost

' Returns the domain interface which is similar to the domain.
' Property Get Domain() As IUnknown

' Returns the .NET system folder.
' Property Get GetCORSystemDirectory() As String

' Returns the version actually loaded.
' Property Get GetCORVersion() As String

' Creates an object from an assembly (including doing the 'unwrap').
' sAssemblyFileName, the filename, full folder is recommended.
' typeName, the classname and if a namespace is used you must pass that as well (like: "MyNamespace.Class1" )
' On success (returns 0 on success) the Dispatch oObject is set with the assembly interface.
' Method CreateObjectByFile( ByVal sAssemblyFileName As String, ByVal typeName As String, ByRef oObject As Dispatch ) As Long


' Tips:
' Many interfaces where not declared in this code, for special use you may need to cast to an interface definition.
' Forget accessing static libraries, since PowerBASIC variable types are so incompatible anyway it's not worth the effort trying to access all objects directly.
' Therefore always use a wrapper assembly which exports the more compatible variabletypes like integers, strings and variants...
' An assembly loaded this way will be loaded as (late bound) dispatch com interface.
' It is possible to define the interface manually though.
' Function overloads are not supported in com, the first function will have it's normal name the 2nd will be numbered like _2 and so on.
' There is no need to do anything special to prepare an assembly for use with PowerBASIC, no comvisible attributes or whatever is needed.
' Assemblies do not need to be modified or registered, just compile and go...

' Example for a simple c# class:
'Using System;
'Using System.Data;
'Using System.IO;
'Using System.Text;
'Using System.Windows.Forms;
'
'/// <summary>
'/// Summary description For Class1
'/// </summary>
'public class Class1
'{
'    public Class1()
'    {
'        //
'        // TODO: Add constructor logic here
'        //
'    }
'
'    public String GetDateTimeStr()
'    {
'        return DateTime.Now.ToString();
'    }
'}

Global DOTNET As DOTNET_IMain

Declare Function DOTNET_CorBindToRuntimeEx( _
      ByVal pwszVersion     As Dword _
    , ByVal pwszBuildFlavor As Dword _
    , ByVal flags           As Dword _
    , ByRef rclsid          As Guid _
    , ByRef riid            As Guid _
    , ByRef ppv             As DOTNET_ICorRuntimeHost _
    ) As Long
Declare Function DOTNET_GetCORSystemDirectory( ByVal pbuffer As Dword, ByVal cchBuffer As Dword, ByRef dwlength As Dword ) As Long
Declare Function DOTNET_GetCORVersion( ByVal pbuffer As Dword, ByVal cchBuffer As Dword, ByRef dwlength As Dword ) As Long
Declare Function DOTNET_System_AppDomain_CreateInstanceFrom ( _
      ByVal pthis         As Dword Ptr _
    , ByVal assemblyFile  As String _
    , ByVal typeName      As String _
    , ByRef pRetVal       As Dispatch _
    ) As Long

'C:\Windows\Microsoft.NET\Framework\v2.0.50727\Mscorlib.tlb
Interface DOTNET_ICorRuntimeHost: Inherit IUnknown

   Method CreateLogicalThreadState() As Long
   Method DeleteLogicalThreadState() As Long
   Method SwitchInLogicalThreadState( ByRef pFiberCookie As Dword ) As Long
   Method SwitchOutLogicalThreadState( ByRef pFiberCookie As Dword ) As Long
   Method LocksHeldByLogicalThread( ByRef pCount As Dword ) As Long
   Method MapFile( ByVal hFile As Dword, ByRef hMapAddress As Dword ) As Long
   Method GetConfiguration( ByRef pConfiguration As IUnknown ) As Long
   Method Start() As Long
   Method Stop() As Long
   Method CreateDomain( _
         ByVal pwzFriendlyName As Dword _
       , ByVal pIdentityArray As IUnknown _
       , ByRef pAppDomain As IUnknown _
       ) As Long
   Method GetDefaultDomain( ByRef pAppDomain As IUnknown ) As Long
   Method EnumDomains( ByRef hEnum As Dword ) As Long
   Method NextDomain( ByVal hEnum As Dword, ByRef pAppDomain As IUnknown ) As Long
   Method CloseEnum( ByVal hEnum As Dword ) As Long
   Method CreateDomainEx( _
         ByVal pwzFriendlyName  As Dword _
       , ByVal pSetup           As IUnknown _
       , ByVal pEvidence        As IUnknown _
       , ByRef pAppDomain       As IUnknown _
       ) As Long
   Method CreateDomainSetup( ByRef pAppDomainSetup As IUnknown ) As Long
   Method CreateEvidence( ByRef pEvidence As IUnknown ) As Long
   Method UnloadDomain( ByVal pAppDomain As IUnknown ) As Long
   Method CurrentDomain( ByRef pAppDomain As IUnknown ) As Long

End Interface

Interface DOTNET_IObjectHandle: Inherit IDispatch

    Property Get ToString <0>() As String
    Method Equals <1610743809>( ByVal obj As Variant ) As Integer
    Method GetHashCode <1610743810>() As Long
    Method GetType <1610743811>() As Dword
    Method GetLifetimeService <1610743812>() As Variant
    Method InitializeLifetimeService <1610743813>() As Variant
    Method CreateObjRef <1610743814>( ByVal requestedType As Dword ) As Dword
    Method Unwrap <1610743815>() As Variant

End Interface

Class clsDOTNET_Main

    Instance MSCoree_hInstance              As Long
    Instance MSCoree_oRuntimeHost           As DOTNET_ICorRuntimeHost
    Instance MSCoree_oDomain                As IUnknown
    Instance MSCoree_pDomain                As Dword Ptr
    Instance MSCoree_pCorBindToRuntimeEx    As Dword
    Instance MSCoree_sGetCORSystemDirectory As String
    Instance MSCoree_sGetCORVersion         As String

    Class Method Destroy
        MSCoree_oRuntimeHost.UnloadDomain( MSCoree_oDomain )
        MSCoree_oDomain = Nothing
        MSCoree_oRuntimeHost.Stop()
        MSCoree_oRuntimeHost = Nothing
        Call FreeLibrary( MSCoree_hInstance ): MSCoree_hInstance = 0
    End Method

    ' Public interface
    Interface DOTNET_IMain: Inherit IUnknown

        Property Get RunTimeHost() As DOTNET_ICorRuntimeHost
            Property = MSCoree_oRuntimeHost
        End Property

        Property Get Domain() As IUnknown
            Property = MSCoree_oDomain
        End Property

        Property Get GetCORSystemDirectory() As String
            Property = MSCoree_sGetCORSystemDirectory
        End Property

        Property Get GetCORVersion() As String
            Property = MSCoree_sGetCORVersion
        End Property

        Method CreateObjectByFile( ByVal sAssemblyFileName As String, ByVal typeName As String, ByRef oObject As Dispatch ) As Long
            Local hr            As Long
            Local oObjectHandle As DOTNET_IObjectHandle
            Local vWrap         As Variant
            sAssemblyFileName = UCode$( sAssemblyFileName )
            typeName = UCode$( typeName )
            Call Dword @@MSCoree_pDomain[38] Using DOTNET_System_AppDomain_CreateInstanceFrom( _
                  MSCoree_pDomain _
                , sAssemblyFileName _
                , typeName _
                , oObjectHandle _
                ) To hr
            If hr = 0 Then
                vWrap = oObjectHandle.UnWrap()
                oObject = vWrap
                vWrap = Empty
                oObjectHandle = Nothing
            End If
            Method = hr
        End Method

    End Interface

    ' Internal use only, don't call yourself.
    Interface DOTNET_IBackDoor As Hidden: Inherit IUnknown

        Method MSCoree_Initialize( _
              ByVal hLib                    As Long _
            , ByVal oHost                   As DOTNET_ICorRuntimeHost _
            , ByVal pCorBindToRuntimeEx     As Dword _
            , ByVal sGetCORSystemDirectory  As String _
            , ByVal sGetCORVersion          As String _
            , ByVal oDomain                 As IUnknown _
            )
            If MSCoree_hInstance Then Exit Method
            MSCoree_hInstance               = hLib
            MSCoree_oRuntimeHost            = oHost
            MSCoree_pCorBindToRuntimeEx     = pCorBindToRuntimeEx
            MSCoree_sGetCORSystemDirectory  = sGetCORSystemDirectory
            MSCoree_sGetCORVersion          = sGetCORVersion
            MSCoree_oDomain                 = oDomain
            MSCoree_pDomain                 = ObjPtr( oDomain )
        End Method

    End Interface

End Class

' Initializes the .NET framework, on succes the global DOTNET com class is present for further use.
' sVersion, the version, for example: v2.0.50727, empty for latest installed version.
' sFlavor, "wks" or "svr", or empty for default.
Function DOTNET_Initialize( ByVal sVersion As String, ByVal sFlavor As String ) As Long

    Local oMain                     As DOTNET_IMain
    Local oMainBackDoor             As DOTNET_IBackDoor
    Local spRuntimeHost             As DOTNET_ICorRuntimeHost
    Local oDomain                   As IUnknown
    Local oDefDomain                As IUnknown
    Local hr                        As Long
    Local CLSID_CorRuntimeHost      As Guid
    Local IID_ICorRuntimeHost       As Guid
    Local IID_AppDomain             As Guid
    Local hLib                      As Long
    Local pCorBindToRuntimeEx       As Dword
    Local pGetCORSystemDirectory    As Dword
    Local pGetCORVersion            As Dword
    Local sGetCORSystemDirectory    As String
    Local sGetCORVersion            As String
    Local dwlength                  As Dword
    Local T                         As String

    If IsObject( DOTNET ) Then Function = 2: Exit Function

    CLSID_CorRuntimeHost = Guid$("{CB2F6723-AB3A-11D2-9C40-00C04FA30A3E}")
    IID_ICorRuntimeHost  = Guid$("{CB2F6722-AB3A-11D2-9C40-00C04FA30A3E}")
    IID_AppDomain        = Guid$("{05F696DC-2B29-3663-AD8B-C4389CF2A713}")

    sFlavor = UCode$( sFlavor )
    sVersion = UCode$( sVersion )

    ' Load the library dynamically, may not be present on all computers.
    hLib = LoadLibrary( "mscoree.dll" )
    Select Case hLib
    Case 0 To 32: Exit Function
    End Select

    ' Fetch addresses to a few functions we need.
    pCorBindToRuntimeEx     = GetProcAddress( hLib, "CorBindToRuntimeEx" )
    pGetCORSystemDirectory  = GetProcAddress( hLib, "GetCORSystemDirectory" )
    pGetCORVersion          = GetProcAddress( hLib, "GetCORVersion" )

    ' Test.
    If pCorBindToRuntimeEx = 0 Or pGetCORSystemDirectory = 0 Or pGetCORVersion = 0 Then
        Call FreeLibrary( hLib )
        Exit Function
    End If

    ' Initializes the run-time.
    Call Dword pCorBindToRuntimeEx Using DOTNET_CorBindToRuntimeEx( _
          IIf&( Len( sVersion ), StrPtr( sVersion ), 0 ) _
        , IIf&( Len( sFlavor ), StrPtr( sFlavor ), 0 ) _
        , 0 _
        , CLSID_CorRuntimeHost _
        , IID_ICorRuntimeHost _
        , spRuntimeHost _
        ) To hr

    If hr <> 0 Then Exit Function

    ' Prefetches some information.
    sGetCORSystemDirectory = UCode$( Space$( %MAX_PATH ) )
    Call Dword pGetCORSystemDirectory Using DOTNET_GetCORSystemDirectory( StrPtr( sGetCORSystemDirectory ), %MAX_PATH, dwlength )
    sGetCORSystemDirectory = Trim$( Parse$( ACode$( sGetCORSystemDirectory ), $Nul, 1 ) )
    sGetCORVersion = UCode$( Space$( %MAX_PATH ) )
    Call Dword pGetCORVersion Using DOTNET_GetCORVersion( StrPtr( sGetCORVersion ), %MAX_PATH, dwlength )
    sGetCORVersion = Trim$( Parse$( ACode$( sGetCORVersion ), $Nul, 1 ) )

    ' Here we go, we need a domain to load assemblies into.
    spRuntimeHost.Start()
    spRuntimeHost.GetDefaultDomain( oDefDomain )
    If IsObject( oDefDomain ) Then oDefDomain.QueryInterface( IID_AppDomain, ByVal VarPtr( oDomain ) )
    oDefDomain = Nothing

    ' Test domain.
    If IsNothing( oDomain ) Then
        spRuntimeHost.Stop()
        Call FreeLibrary( hLib )
        Exit Function
    End If

    ' Everything seems ok, prepare the global PowerBASIC com class to work with.
    ' This will unload the .NET framework when going out of scope (on app termination).
    DOTNET = Class "clsDOTNET_Main"

    ' Pass settings on via its backdoor interface.
    oMainBackDoor = DOTNET
    oMainBackDoor.MSCoree_Initialize( _
          hLib _
        , spRuntimeHost _
        , pCorBindToRuntimeEx _
        , sGetCORSystemDirectory _
        , sGetCORVersion _
        , oDomain _
        )

    Function = 1

End Function
